import React, { Component } from 'react';
import { Machine } from 'xstate';
import { interpret } from 'xstate/lib/interpreter';
import styled from 'styled-components';
import { assign } from 'xstate/lib/actions';

const addressBarMachine = Machine(
  {
    id: 'addressBar',
    initial: 'idle',
    context: {
      initial: '',
      location: '',
      list: [],
      selected: undefined
    },
    on: {
      BLUR: 'idle'
    },
    states: {
      idle: {
        on: {
          CLICK: { target: 'active' }
        }
      },
      active: {
        initial: 'closed',
        states: {
          closed: {
            onEntry: ['selectAll'],
            on: {
              'KEY.DOWN': { target: 'selecting' },
              'KEY.ESC': { actions: ['setInitial', 'selectAll'] },
              KEY: 'selecting'
            }
          },
          opened: {
            initial: 'autocompleted',
            states: {
              autocompleted: {
                on: {
                  BACKSPACE: {
                    target: 'not autocompleted',
                    actions: 'remove autocompleted text'
                  }
                }
              },
              'not autocompleted': {
                on: {
                  BACKSPACE: {
                    target: 'not autocompleted',
                    actions: 'update text'
                  },
                  'KEY.UP': '#addressBar.selecting.withSearch'
                }
              }
            },
            on: {
              'KEY.UP': 'selecting',
              'KEY.DOWN': {
                target: 'selecting',
                actions: 'selectNextItem'
              },
              KEY: {
                target: 'autocompleted',
                actions: 'update text'
              }
            }
          },
          selecting: {
            initial: 'choose',
            states: {
              choose: {
                on: {
                  '': [
                    {
                      target: 'autoCompleting',
                      cond: ctx => ctx.selected.indexOf(ctx.location) === 0
                    },
                    { target: 'typing' }
                  ]
                }
              },
              autoCompleting: {
                on: {
                  ESC: 'opened.autocompleted'
                }
              },
              typing: {
                onEntry: ['update', 'populateList']
              }
            },
            on: {
              'KEY.DOWN': {
                actions: ['selectNextItem', 'setSelected']
              },
              'KEY.UP': {
                actions: ['selectPrevItem', 'setSelected']
              },
              'KEY.ESC': {
                target: 'closed'
              }
            }
          }
        },
        on: {
          'KEY.ESC': '.closed',
          KEY: [
            {
              target: '.closed',
              cond: (_, e) => e.value === '',
              actions: ['update']
            },
            {
              actions: ['update', 'populateList']
            }
          ]
        }
      }
    }
  },
  {
    actions: {
      selectNextItem: assign({
        selected: ctx => {
          return ctx.list[ctx.list.indexOf(ctx.selected) + 1] || ctx.selected;
        }
      }),
      selectPrevItem: assign({
        selected: ctx => {
          return ctx.list[ctx.list.indexOf(ctx.selected) - 1] || ctx.selected;
        }
      }),
      selectFirstItem: assign({
        selected: ctx => ctx.list[0]
      }),
      setSelected: assign({
        location: ctx => ctx.selected
      }),
      setInitial: assign({
        location: ctx => ctx.initial
      }),
      update: assign({
        location: (_, e) => e.value
      }),
      populateList: assign({
        list: ctx => {
          console.log(ctx.location);
          return ctx.dataSet
            .filter(item => item.includes(ctx.location, 0))
            .concat('search');
        }
      })
    }
  }
);

const autoCompleteList = ['one', 'two', 'three', 'four', 'testing'];

export class AddressBar extends React.Component {
  state = {
    current: addressBarMachine.initialState
  };
  interpreter = interpret(
    addressBarMachine.withConfig(
      {
        actions: {
          selectAll: () =>
            setTimeout(
              () => this.searchRef.current && this.searchRef.current.select()
            )
        }
      },
      {
        initial: 'testing',
        location: 'testing',
        dataSet: autoCompleteList,
        list: [],
        selected: autoCompleteList[0]
      }
    )
  ).onTransition(current => {
    this.setState({ current });
  });
  searchRef = React.createRef();
  componentDidMount() {
    this.interpreter.start();
  }
  reset() {
    // this.interpreter.stop();
    this.interpreter.start();
  }
  render() {
    const { current } = this.state;
    const { send } = this.interpreter;

    console.log(current.matches('active.opened'));

    return (
      <div>
        <input
          type="search"
          onClick={_ => send('CLICK')}
          onBlur={_ => send('BLUR')}
          onKeyDown={e => {
            console.log(e.key);
            switch (e.key) {
              case 'ArrowDown': {
                e.preventDefault();
                return send('KEY.DOWN');
              }
              case 'ArrowUp': {
                e.preventDefault();
                return send('KEY.UP');
              }
              case 'Escape': {
                e.preventDefault();
                return send('KEY.ESC');
              }
              default: {
              }
            }
          }}
          onChange={e => {
            console.log(e.target.value, '<<<');
            send({ type: 'KEY', value: e.target.value });
          }}
          ref={this.searchRef}
          value={current.context.location}
        />
        {current.matches('active.opened') ||
        current.matches('active.selecting') ? (
          <ul>
            {current.context.list.map(item => {
              return (
                <li
                  key={item}
                  style={{
                    background:
                      current.context.selected === item
                        ? 'rgba(255, 255, 255, 0.5)'
                        : 'transparent'
                  }}
                >
                  {item}
                </li>
              );
            })}
          </ul>
        ) : null}
        <pre>{JSON.stringify(current.value)}</pre>
        <button onClick={_ => this.reset()}>reset</button>
      </div>
    );
  }
}
